/*
 * file: pv_event.c
 * This file is part of the libpv library.
 *
 * Copyright 2005 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 *
 * PURPOSE:
 * Provides event notification for resource changes
 * (change notification)
 */

#include <libpv_config.h>

#include <errno.h>
#include <getopt.h>
#include <libgen.h>
#include <stdlib.h>
#include <string.h>
#include <mqueue.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/stat.h>

#include <pv_types.h>
#include <pv_lock.h>
#include <pv_ev.h>
#include <pv_register.h>

#include <unistd.h>

/**
 * pv_event_init_handling - Intialize the event handler for this process context
 * @engine: Current engine to work with
 * @store_filename: Filename of the backing store
 *
 * Description: This creates a resource to deliver events to waiting clients.
 * It creates it local for the calling process and save it into the given
 * engine.
 * Current mechanism: Two message queues are used for this purpose.
 * On in server's direction, one to receive change notifications.
 * Thats mostly all.
 *
 * Return value: 0 if successfull, -EINVAL if you try to init an engine more than once,
 * some other values if it wasn't possible to create the message queues
 **/
int pv_event_init_handling(struct pv_engine *engine,const char *store_filename)
{
	int	err;
	char	name[30];
	struct mq_attr mq_attr;
	struct pve_message message;
	/*
	 * only to be sure...a little bit paranoia
	 */
	if (engine->mqdes != (mqd_t)-1) {
		dprintf(LEVEL_ERROR, "(%s) Connection to server seems already existing. Do you call me more than once?\n",__FUNCTION__);
		return(-EINVAL);
	}
	if (engine->mq_notify != (mqd_t)-1) {
		dprintf(LEVEL_ERROR, "(%s) Connection from server seems already existing. Do you call me more than once?\n",__FUNCTION__);
		return(-EINVAL);
	}
	/*
	 * we play around with global (shared) resources.
	 * Due to the fact we have to initialize them if
	 * we are the first one here, its racy to do that.
	 * So lock the engine, to be lonely here
	 */
	pv_lock_lock(engine);

	engine->event_server_key=ftok(store_filename, 43);
	/*
	 * Open the connection to the server
	 * Must exist to support event notification!
	 * If there is no server, event notification is
	 * disabled
	 */
	snprintf(name,sizeof(name),QUEUE_NAME,engine->event_server_key);
	engine->mqdes=mq_open(name,O_WRONLY);
	if (engine->mqdes == (mqd_t)-1) {
		err=-errno;
		dprintf(LEVEL_INFO,"(%s) cannot open message queue %s to server: %s. Disabling "
			"event notification\n",__FUNCTION__,name,strerror(errno));
		goto error_1;
	}
	/*
	 * Create the notify queue
	 */
	snprintf(name,sizeof(name),QUEUE_CLIENT,getpid());
	/*
	 * setup attributes for this queue:
	 */
	mq_attr.mq_flags=0;	/* this queue blocks */
	mq_attr.mq_maxmsg=10;	/* FIXME: how to calculate a good size? */
	mq_attr.mq_msgsize=sizeof(struct pve_message);
	mq_attr.mq_curmsgs=0;	/* nothing in it yet */

	engine->mq_notify = mq_open(name,O_RDWR | O_CREAT,QUEUE_PERM,&mq_attr);

	if (engine->mq_notify == (mqd_t)-1) {
		err=-errno;
		dprintf(LEVEL_CRIT,"(%s) cannot create notify queue %s: %s\n",
			__FUNCTION__,name,strerror(errno));
		mq_unlink(name);	/* FIXME: Why it sometimes fails without that? */
		goto error_2;
	}
	engine->mq_name=strndup(name,sizeof(name));
	pv_lock_unlock(engine);
	/*
	 * now notify the server about our existance
	 */
	message.type=PVE_ADD_CLIENT;
	message.pid=getpid();	/* add me! */
	mq_send(engine->mqdes,(char*)&message,sizeof(message),0);
	return(0);

error_2:
	mq_close(engine->mqdes);
	engine->mqdes=(mqd_t)-1;
error_1:
	pv_lock_unlock(engine);
	return(err);
}

/**
 * pv_event_exit_handling - Remove all resources for this purpose
 * @engine:  Current engine to work with
 *
 * Description: Remove all claimed resources. Most important part is to
 * unlink the notification message queue.
 *
 * Return value: 0 on success, or return values from mq_close. Its only to inform,
 * the functions runs completly.
 **/
int pv_event_exit_handling(struct pv_engine *engine)
{
	int err=0;
	struct pve_message message;
	/*
	 * Notify the server that we are going do leave
	 */
	message.type=PVE_REMOVE_CLIENT;
	message.pid=getpid();	/* FIXME: ensure its the same when register */
	mq_send(engine->mqdes,(char*)&message,sizeof(message),0);

	if (engine->mqdes != (mqd_t)-1) {	/* do we claim a message queue to the server? */
		dprintf(LEVEL_INFO, "(%s) closing message queue to server.\n",__FUNCTION__);
		err=mq_close(engine->mqdes);
		if (err != 0)
			dprintf(LEVEL_CRIT,"(%s) cannot close message queue to server. err %s\n",
				__FUNCTION__,strerror(errno));
		engine->mqdes = (mqd_t)-1;
	}

	if (engine->mq_notify != (mqd_t)-1) {	/* do we claim a notify queue? */
		dprintf(LEVEL_INFO, "(%s) closing notify queue.\n",__FUNCTION__);
		err=mq_close(engine->mq_notify);
		if (err != 0)
			dprintf(LEVEL_CRIT,"(%s) cannot close notify queue. err %s\n",
				__FUNCTION__,strerror(errno));
		/*
		 * This unlink may fail, if the server was faster in doing so.
		 * Thats why we ignore the returned value.
		 * See remove_client() in pv_eventd.c
		 */
		mq_unlink(engine->mq_name);
		engine->mq_notify = (mqd_t)-1;
		free(engine->mq_name);
		engine->mq_name=NULL;
	}

	return(err);
}

/**
 * pv_event_register - Connect a notification to a process variable
 * @engine_anon: Current engine to work with
 * @pv_var: Process variable
 *
 * Description: Add ourself to the list of notified clients for this variable.
 * Its handled on server side.
 *
 * Return value: 0 if successfull, -EINVAL if the process variable seems invalid
 **/
int pv_event_register(const pv_engine_t *engine_anon,const void *pv_var)
{
	struct pv_engine	*engine = (struct pv_engine *)engine_anon;
	off_t offset;
	struct pve_message message;

/* FIXME: Maybe a more robust test for variable validation! */

	if ((pv_var == NULL)  || (engine->mqdes == -1))
		return(-EINVAL);

	offset=abs_addr_to_offset(engine,pv_var);
	dprintf(LEVEL_INFO, "(%s) Register var %p (=%u).\n",__FUNCTION__,pv_var,(unsigned int)offset);

	message.type=PVE_ADD_PV;
	message.pid=getpid();	/* FIXME: ensure its the same when register */
	message.message.offset=offset;
	mq_send(engine->mqdes,(char*)&message,sizeof(message),0);

	return(0);
}

/**
 * pv_event_deregister - Disconnect a notification from a process variable
 * @engine_anon: Current engine to work with
 * @pv: Process variable
 *
 * Description: Delete ourself from the list of notified clients for this variable
 * Its handled on server side.
 *
 * Return value: 0 if successfull, -EINVAL if the process variable seems invalid
 * or no event daemon is available
 **/
int pv_event_deregister(const pv_engine_t *engine_anon,const void *pv_var)
{
	struct pv_engine	*engine = (struct pv_engine *)engine_anon;
	off_t offset;
	struct pve_message message;

/* FIXME: Maybe a more robust test for variable validation! */

	if ((pv_var == NULL) || (engine->mqdes == -1))
		return(-EINVAL);

	offset=abs_addr_to_offset(engine,pv_var);
	dprintf(LEVEL_INFO, "(%s) Deregister var %p (=%u).\n",__FUNCTION__,pv_var,(unsigned int)offset);

	message.type=PVE_REMOVE_PV;
	message.pid=getpid();	/* FIXME: ensure its the same when register */
	message.message.offset=offset;
	mq_send(engine->mqdes,(char*)&message,sizeof(message),0);

	return(0);
}

/**
 * pv_event_await_syncronous - Block until a notification arrives
 * @engine_anon: Current engine to work with
 * @pv_var: This outputs the process variable a notification for arrives
 *
 * Description: This function blocks until at least one notification was
 * received. It returns exactly one process variable per call.
 * So it must call n times to receive n notifications.
 *
 * Return value: 0 if it returns a process variable in @pv_var
 * successfully or somthing that mq_receive returns on failure
 * It returns -EINVAL if there is no event daemon available
 **/

int pv_event_await_syncronous(const pv_engine_t *engine_anon,void **pv_var)
{
	struct pv_engine	*engine = (struct pv_engine *)engine_anon;
	ssize_t length;
	int err=0;
	struct pve_message message;

	if (engine->mqdes == -1) {
		*pv_var=NULL;
		return(-EINVAL);
	}

	length=mq_receive(engine->mq_notify,(char*)&message,sizeof(message),NULL);
	if (length == (ssize_t)-1) {
		err=-errno;
		dprintf(LEVEL_CRIT,"(%s) message queue error. Sizing mismatch: Should be %u, is %zd: %s\n",
			__FUNCTION__,(unsigned int)sizeof(message),length,strerror(errno));
	}
	else {
		/* FIXME: should check if this offset belongs to a valid pv */
		*pv_var=engine->store.map + message.message.offset;
	}

	return(err);
}

/**
 * pv_event_generate - Generate a notification for a process variable
 * @engine_anon: Current engine to work with
 * @pv_var: Generate event for exactly this process variable
 *
 * Description: pv_event_generate notifies the server about the
 * change. The server disributes this notification to all
 * registered clients for this process variable.
 *
 * Return value: 0 on success, -EINVAL if @pv_var is invalid or no event
 * daemon is available
 **/
int pv_event_generate(const pv_engine_t *engine_anon,const void *pv_var)
{
	struct pv_engine	*engine = (struct pv_engine *)engine_anon;
	struct pve_message message;

	if ((pv_var == NULL) || (engine->mqdes == -1))
		return(-EINVAL);

	message.type=PVE_CHANGED;
	message.pid=getpid();	/* FIXME: ensure its the same when register */
	message.message.offset=abs_addr_to_offset(engine,pv_var);

	dprintf(LEVEL_INFO, "(%s) Var %u changed.\n",__FUNCTION__,
		(unsigned int)message.message.offset);

	mq_send(engine->mqdes,(char*)&message,sizeof(message),0);

	return(0);
}

/* end of file pv_event.c */
