/*
 * file: pv_eventd.c
 * This file is part of the libpv library.
 *
 * Copyright 2005 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 *
 * PURPOSE:
 * This is the event daemon (change notification server)-
 * It works on a per process image basis.
 *
 * This daemon handles all other processes
 * (here called clients) to work with this
 * process image.
 */

#include <libpv_config.h>

#include <errno.h>
#include <getopt.h>
#include <libgen.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <sys/ipc.h>	/* for ftok() */

#include <pv_types.h>
#include <pv_lock.h>

#include <unistd.h>

struct registered_client {
	pid_t pid;
	mqd_t msgq;
	struct registered_client *next;
};

struct client_list {
	struct registered_client *head;
	struct registered_client *tail;
	struct registered_client *cur;
};

struct pve {
	off_t	offset;
	struct client_list	*registered_clients;	/* list of clients registered for this variable */
	struct pve *next;
};

struct pve_list {
	struct pve *head;
	struct pve *tail;
	struct pve *cur;
};

/* this list holds data for any connected client */
static struct client_list *client_list;
/* This list manages the registrations */
static struct pve_list *pve_list;

/*
 * List's descriptions:
 *
 * client_list (stores all currently registered clients)
 *   ---> struct registered client ---> struct registered client ---> struct registered client
 *
 * pve_list (stores all currently administered process variables)
 *   ---> struct pve ---> struct pve ---> struct pve ---> struct pve
 *
 * per "struct pve" (registered clients to notify when changed)
 *   ---> struct registered client  ---> struct registered client  ---> struct registered client
 */

/* engine we control */
static struct pv_engine pv_engine;

/**
 * find_client - search for a client in the given list
 * @list: The list to search in
 * @pid: The client to search for
 *
 * Description: Helper function
 *
 * Return value: Pointer to list element the requested client is in, NULL otherwise
 **/
static struct client_list *find_client(struct client_list *list,const pid_t pid)
{
	int finished=0;
	struct registered_client *client;

	if (!is_empty(list)) {
		for (goto_head(&list); !finished; goto_next(&list)) {
			client = (struct registered_client*)list_get_element_data(list);
			if (is_last(list))
				finished = 1;
			if (client->pid == pid)
				return(list);
		}
	}

	return(NULL);	/* unknown client! */
}

/**
 * find_pv - search for a variable in the given list
 * @list: The list to search in
 * @offset: Variable's offset
 *
 * Description: Helper function
 *
 * Return value: Pointer to list element the requested variable is in, NULL otherwise
 **/

static struct pve_list *find_pv(struct pve_list *list,const off_t offset)
{
	int finished=0;
	struct pve *pve;

	if (!is_empty(list)) {
		for (goto_head(&list); !finished; goto_next(&list)) {
			pve = (struct pve*)list_get_element_data(list);
			if (is_last(list))
				finished = 1;
			if (pve->offset == offset)
				return(list);
		}
	}

	return(NULL);	/* unknown process variable! */
}

/**
 * pv_create_eye - create an eye to catch all relevant client data
 * @engine: Engine to work for
 * @filename: the persistant image file
 *
 * Description:
 *
 * Return value:
 **/
static int pv_create_eye(struct pv_engine *engine,const char *filename)
{
	int	err;
	char	name[30];
	struct mq_attr mq_attr;

	dprintf(LEVEL_INFO, "(%s) Create server's eye.\n",__FUNCTION__);
	/*
	 * only to be sure...a little bit paranoia
	 */
	if (engine->mqdes != (mqd_t)-1) {
		dprintf(LEVEL_ERROR, "(%s) Message queue already exist. Do you call me more than once?\n",__FUNCTION__);
		return(-1);
	}
	/*
	 * Create an unique ID to ensure every client
	 * can connect us, if it creates a key in the
	 * same way
	 */
	engine->event_server_key=ftok(filename, 43);
	/*
	 * Create our own message queue
	 * Attention: The name of the queue must be unique.
	 */
	snprintf(name,sizeof(name),QUEUE_NAME,engine->event_server_key);
	/*
	 * setup attributes for this queue:
	 */
	mq_attr.mq_flags=0;	/* this queue blocks */
	mq_attr.mq_maxmsg=10;	/* FIXME: how to calculate a good size? And what is allowed? 100 fails! */
	mq_attr.mq_msgsize=sizeof(struct pve_message);
	mq_attr.mq_curmsgs=0;	/* nothing in it yet */
	/*
	 * Open it exclusive, we must be the first for this process image
	 */
	engine->mqdes = mq_open(name,
			O_RDWR | O_CREAT | O_EXCL,		/* RW to be able to send myself messages */
			QUEUE_PERM,
			&mq_attr);

	if (engine->mqdes == (mqd_t)-1) {
		err=-errno;
		dprintf(LEVEL_CRIT,"(%s) cannot create message queue %s: %s\n",
			__FUNCTION__,name,strerror(errno));
		mq_unlink(name);	/* FIXME: Why it sometimes fails without that? */
		return(err);
	}
	engine->mq_name=strndup(name,sizeof(name));

	return(0);
}

/**
 * pv_remove_eye - remove claimed resources
 * @engine: Engine to work for
 *
 * Description: Delete and unlink the messag queue. Don't forget to
 * unlink. If you do so, next time it fails to open this queue.
 *
 * Return value: 0 on success
 **/
static int pv_remove_eye(struct pv_engine *engine)
{
	int err;

	dprintf(LEVEL_INFO, "(%s) Remove server's eye.\n",__FUNCTION__);

	if (engine->mqdes != (mqd_t)-1) {
		dprintf(LEVEL_INFO, "(%s) close message queue.\n",__FUNCTION__);
		err=mq_close(engine->mqdes);
		if (err != 0)
		dprintf(LEVEL_CRIT,"(%s) cannot close message queue handle. err %s\n",
			__FUNCTION__,strerror(errno));
		dprintf(LEVEL_INFO, "(%s) unlink message queue.\n",__FUNCTION__);
		mq_unlink(engine->mq_name);
		engine->mqdes = (mqd_t)-1;
		free(engine->mq_name);
		engine->mq_name=NULL;
	}
	return(0);
}

/**
 * pv_setup_static_data - create and setup static data to do the work here
 * @engine: Engine to work for
 *
 * Description: Setup a local list of process variables to handle future
 * requests for notfication.
 * Caution: Uses global pointers pve_list and client_list
 *
 * Return value: 0 on success
 **/
static int pv_setup_static_data(struct pv_engine *engine)
{
	int	finished=0;
	struct pv	*pv;
	struct pve	pve;

	dprintf(LEVEL_INFO, "(%s) Create static data.\n",__FUNCTION__);
	/*
	 * setup the list of known pv
	 */
	list_init(pve_list);
	for (goto_head(&engine->pv_list); !finished; goto_next(&engine->pv_list)) {
		pv = (struct pv *)list_get_element_data(engine->pv_list);
		if (is_last(engine->pv_list))
			finished = 1;
		pve.offset=pv->offset;
		list_init(pve.registered_clients);	/* no one registered yet */
		list_add_tail(pve_list,(void *)&pve,sizeof(struct pve));
	}
	/*
	 * setup the list of known clients
	 */
	list_init(client_list);
	return(0);
}

/**
 * pv_remove_static_data - remove all claimed resources
 * @pv_engine:
 *
 * Description: Remove all local lists
 * Caution: Uses global pointers pve_list and client_list
 *
 * FIXME: Use remove_client() instead of doing it here again
 *
 * Return value:  0 on success
 **/
static int pv_remove_static_data(struct pv_engine *pv_engine)
{
	int	finished=0;
	struct pve	*pve;
	struct registered_client *client;
	char name[30];

	dprintf(LEVEL_INFO, "(%s) Remove static data.\n",__FUNCTION__);

	if (!is_empty(pve_list)) {	/* something registered? */
		for (goto_head(&pve_list); !finished; goto_next(&pve_list)) {
			pve = (struct pve*)list_get_element_data(pve_list);
			if (is_last(pve_list))
				finished = 1;
			list_del(pve->registered_clients);
			list_del_element(pve_list);
		}
	}
	else
		list_del(pve_list);

	if (!is_empty(client_list)) {	/* someone registered? */
		finished=0;
		for (goto_head(&client_list); !finished; goto_next(&client_list)){
			client=(struct registered_client*)list_get_element_data(client_list);
			if (is_last(client_list))
				finished = 1;
			/*
			 * Close and remove the message queue. Maybe the client was killed,
			 * there is no one to remove it anymore
			 */
			mq_close(client->msgq);
			snprintf(name,sizeof(name),QUEUE_CLIENT,client->pid);
			mq_unlink(name);
			list_del_element(client_list);
		}
	}
	else
		list_del(client_list);

	return(0);
}

/**
 * notify_this_client - notify one client about a variable change
 * @client: Client to notify
 * @offset: Offset of this variable
 *
 * Description: To simplify the mechanism to notify someone
 * this is an separat function. Let the compiler do some
 * optimization, static is a good candidate to do so.
 *
 * Return value: 0 on success, otherwise -(return code)
 * from mq_send-call if it fails
 **/
static int notify_this_client(struct registered_client *client,off_t offset)
{
	struct pve_message message;
	int err=0;

	dprintf(LEVEL_INFO, "(%s) Notify client %d.\n",__FUNCTION__,client->pid);

	message.type=PVE_CHANGED;
	message.pid=client->pid;	/* only to be sure */
	message.message.offset=offset;

	if (mq_send(client->msgq,(char*)&message,sizeof(message),0) == -1) {
		if (errno == EBADF)	/* not a valid message queue? */
			; /* FIXME: delete any registration of this client! It seems dead */
		else {
			dprintf(LEVEL_CRIT,"(%s) message queue error: %s",
				__FUNCTION__,strerror(errno));
				err=-errno;
		}
	}
	return(err);
}

/**
 * pv_notify_clients - Notify all clients registered for this variable
 * @pves: List of variables and registered clients
 * @offset: Offset of this variable
 *
 * Description:
 *
 * Return value: 0 on success
 **/
static int pv_notify_clients(struct pve_list **pves,off_t offset)
{
	struct pve	*pve;
	struct pve_list *pve_l;
	struct client_list *rc_list;
	struct registered_client *rc;
	int finished=0;

	dprintf(LEVEL_INFO, "(%s) Notify about change for %u.\n",__FUNCTION__,(unsigned int)offset);
	/*
	 * search for the process variable
	 */
	if ((pve_l=find_pv(*pves,offset)) == NULL) {
		dprintf(LEVEL_WARN,"(%s) Unknown variable (%u) to notify!",
				__FUNCTION__,(unsigned int)offset);	/* FIXME: Should this better be an error? */
		return(0);	/* ignore */
	}

	pve = (struct pve*)list_get_element_data(pve_l);
	if (!is_empty(pve->registered_clients)) {	/* someone registered? */
		/*
		 * walk through the list for registerd clients
		 */
		for (rc_list=pve->registered_clients,goto_head(&rc_list);
			!finished; goto_next(&rc_list)) {
			rc = (struct registered_client*)list_get_element_data(rc_list);
			if (is_last(rc_list))
				finished = 1;
			notify_this_client(rc,offset);
		}
	}
	return(0);
}

/**
 * register_pv - register notification for a client on a process variable
 * @pves: List of variables to register
 * @pid: client to register
 * @offset: offset of the process variable to register for
 *
 * Description:
 *
 * Return value: 0 on success, -EINVAL if client and/or variable are unknown
 **/
static int register_pv(struct pve_list **pves,struct client_list **clients,const pid_t pid,const off_t offset)
{
	struct pve_list *pve_l;
	struct client_list *client_l,*registered_l;
	struct registered_client *client;
	struct pve	*pve;

	dprintf(LEVEL_INFO, "(%s) Register client %d to variable %u.\n",__FUNCTION__,
		pid,(unsigned int)offset);
	if ((client_l=find_client(*clients,pid)) == NULL)
		return(-EINVAL);	/* client not registered yet */
	if ((pve_l=find_pv(*pves,offset)) == NULL)
		return(-EINVAL);	/* unknown process variable */
	/*
	 * Client is valid and requested process variable also.
	 * Add this client to the list of variable's registered clients
	 */
	client=(struct registered_client*)list_get_element_data(client_l);
	pve=(struct pve*)list_get_element_data(pve_l);

	if ((registered_l=find_client(pve->registered_clients,client->pid)) != NULL) {
		dprintf(LEVEL_WARN,"(%s) Client %d already registered to var %u\n",
				__FUNCTION__,pid,(unsigned int)offset);
		return(0);	/* this is not an error */
	}
	list_add_tail(pve->registered_clients,
		(void*)client,sizeof(struct registered_client));

	return(0);
}

/**
 * deregister_pv - deregister notification from a process variable
 * @pves: List of variables
 * @pid: client to deregister
 * @offset: offset of the process variable to deregister
 *
 * Description:
 *
 * Return value: 0 on success, -EINVAL if no process variable
 **/
static int deregister_pv(struct pve_list **pves,const pid_t pid,const off_t offset)
{
	struct pve	*pve;
	struct pve_list *pve_l;
	struct client_list *registered_l;

	dprintf(LEVEL_INFO, "(%s) Deregister client %d from variable %u...\n",
		__FUNCTION__,pid,(unsigned int)offset);
	/*
	 * search for the process variable
	 */
	if ((pve_l=find_pv(*pves,offset)) == NULL)
		return(-EINVAL);	/* unknown process variable */
	pve=(struct pve*)list_get_element_data(pve_l);
	/*
	 * find the client in the list and remove it
	 */
	if ((registered_l=find_client(pve->registered_clients,pid)) != NULL)
		list_del_element(registered_l);

	return(0);
}

/**
 * add_client - add a client to handle here
 * @clients: List to add the given client into
 * @pid: This client should be added
 *
 * Description:
 * Caution: Modifies global pointer client_list
 *
 * Return value: 0 on success, otherwise -(return code)
 * from mq_open-call if it fails
 **/
static int add_client(struct client_list **clients,const pid_t pid)
{
	struct registered_client client;
	struct client_list *client_l;
	char name[30];

	dprintf(LEVEL_INFO, "(%s) Add a connection for client %d.\n",__FUNCTION__,pid);

	if ((client_l=find_client(*clients,pid)) != NULL) {
		dprintf(LEVEL_WARN,"(%s) Client %d allready registered!",
				__FUNCTION__,pid);
		return(0);
	}

	client.pid=pid;
	snprintf(name,sizeof(name),QUEUE_CLIENT,pid);
	/*
	 * Open a queue back to the client. This queue
	 * must already exist!
	 */
	client.msgq = mq_open(name,O_WRONLY);

	if (client.msgq == (mqd_t)-1) {
		dprintf(LEVEL_CRIT,"(%s) message queue error: %s",
			__FUNCTION__,strerror(errno));
		return(-errno);
	}

	list_add_tail(*clients,(void *)&client,
		sizeof(struct registered_client));

	return(0);
}

/**
 * remove_client - remove a client from handling here
 * @clients: Client list to remove the given client from
 * @pves: PV list to remove the given client from
 * @pid: This client should be removed
 *
 * Description:
 * Caution: Modifies global pointer client_list
 *
 * Return value: 0 on success
 **/
static int remove_client(struct client_list **clients,struct pve_list **pves,const pid_t pid)
{
	struct registered_client *client;
	struct client_list *client_l;
	int finished=0;
	struct pve *pve;
	char name[30];

	dprintf(LEVEL_INFO, "(%s) Remove a connection to client %d.\n",__FUNCTION__,pid);

	if ((client_l=find_client(*clients,pid)) == NULL)
		dprintf(LEVEL_WARN,"(%s) Unknown client %d to remove!\n",
				  __FUNCTION__,pid);
	else {
		/*
		 * step through all process variables and
		 * deregister this client
		 */
		if (!is_empty(*pves)) {
			for (goto_head(pves); !finished; goto_next(pves)) {
				pve = (struct pve*)list_get_element_data(*pves);
				if (is_last(*pves))
					finished = 1;
				deregister_pv(pves,pid,pve->offset);
			}
		}
		/*
		 * Close our side if this message queue and also remove it.
		 * Maybe the client was killed, so we here are responsible
		 * to free this kind of system resource
		 */
		client=(struct registered_client*)list_get_element_data(client_l);
		if (mq_close(client->msgq) != 0) {
			dprintf(LEVEL_WARN,"(%s) Message queue error: %s",
				__FUNCTION__,strerror(errno));
		}
		snprintf(name,sizeof(name),QUEUE_CLIENT,pid);
		/*
		 * This unlink may fail if the client was faster doing so.
		 * Thats why we ignore the returned value.
		 * See pv_event_exit_handling() in pv_event.c
		 */
		mq_unlink(name);
		list_del_element(client_l);	/* remove this one */
	}

	return(0);
}

/**
 * do_the_job - loops endless to handle all client requests
 * @engine: Engine to work for
 * @clients: List of clients -> access to the anchor pointer is required!
 * @pves: List of process variables -> access to the anchor pointer is required!
 *
 * Description: Handles all the messages and interacts with the clients
 *
 * Return value:
 **/
static int do_the_job(struct pv_engine *pv_engine,struct client_list **clients,struct pve_list **pves)
{
	ssize_t length;
	int err=0;
	struct pve_message message;

	dprintf(LEVEL_INFO, "(%s) Waiting for client messages.\n",__FUNCTION__);
	while(1) {
		length=mq_receive(pv_engine->mqdes,(char*)&message,sizeof(message),NULL);
		if (length == (ssize_t)-1) {
			err=-errno;
			dprintf(LEVEL_CRIT,"(%s) message queue error. Sizing mismatch: Should be %u, is %zd: %s",
				__FUNCTION__,(unsigned int)sizeof(message),length,strerror(errno));
			return(err);
		}

		switch (message.type) {
		case PVE_ADD_CLIENT:
			add_client(clients,message.pid);
			break;

		case PVE_REMOVE_CLIENT:
			remove_client(clients,pves,message.pid);
			break;

		case PVE_ADD_PV:
			register_pv(pves,clients,message.pid,message.message.offset);
			break;

		case PVE_REMOVE_PV:
			deregister_pv(pves,message.pid,message.message.offset);
			break;

		case PVE_CHANGED:
			pv_notify_clients(pves,message.message.offset);
			break;

		case PVE_SERVER_QUIT:	/* FIXME: Who should be able do send this kind of message? */
			dprintf(LEVEL_NOTICE, "(%s) Server receives QUIT message from client %d.\n",
			 	__FUNCTION__,message.pid);
			return(0);

		default:
			dprintf(LEVEL_ERROR, "(%s) Unknown message (%d) from client %d received. Ignore.\n",
				__FUNCTION__,message.type,message.pid);
		}
	}
}

/**
 * signalhandler - inform all other clients that we are going to fade away..
 * @sig: Catched signal
 **/
void signalhandler (int sig)
{
	struct pve_message message;

	if (sig == SIGPIPE)
		return;
	if (sig == SIGHUP) {
		fprintf(stderr, "SIGHUP received\n");
	}

	if (sig == SIGTERM || sig == SIGINT) {
		fprintf(stderr, "received SIGTERM or SIGINT, shutting down\n");

		message.type=PVE_SERVER_QUIT;
		message.pid=getpid();

		mq_send(pv_engine.mqdes,(char*)&message,sizeof(message),0);
	}
}

int main(int argc, char *argv[])
{
	pv_engine_attr_t	attr;
	int			err;

	err = pv_engine_attr_init_from_cmdline(&attr, argc, argv);
	if (err)
		exit (EXIT_FAILURE);

	attr.is_server=1;

	err = pv_init((pv_engine_t*)&pv_engine, &attr);
	if (err)
		exit (EXIT_FAILURE);
	/*
	 * Now we are connected to this process image.
	 * We lock the engine until we will created the
	 * message queue and the static data successfully.
	 * Then it is save to receive messages from new clients
	 */
	pv_lock_lock(&pv_engine);
	/*
	 * Setup all relevant static data, create a
	 * message queue and wait for additional
	 * connections by new clients
	 */
	err=pv_setup_static_data(&pv_engine);
	if (err)
		goto error_1;

	err=pv_create_eye(&pv_engine,attr.store_filename);
	if (err)
		goto error_2;
	/*
	 * Now we are ready to do our supervisor job
	 */
	pv_lock_unlock(&pv_engine);

	/*
	 * Setup a signal handler for SIGINT
	 */
	signal(SIGINT, signalhandler);
	signal(SIGTERM, signalhandler);

	/*
	 * do it
	 */
	do_the_job(&pv_engine,&client_list,&pve_list);

	pv_remove_eye(&pv_engine);
	pv_destroy((pv_engine_t*)&pv_engine);
	pv_remove_static_data(&pv_engine);
	return(EXIT_SUCCESS);

error_2:
	pv_remove_static_data(&pv_engine);
error_1:
	pv_lock_unlock(&pv_engine);
	pv_destroy((pv_engine_t*)&pv_engine);
	return(err);
}

/* end of file pv_eventd.c */
